# -*- coding: utf-8 -*-
"""
Created on Tue Sep 22 08:17:44 2015

@author: Hunter Cole Davis (hdavis AT caltech DOT edu)
@Citation: Mukherjee, A., Wu D, Davis HC, Shapiro MG. Non-invasive imaging using reporter genes altering cellular water permeability. Nat. Commun. 7, 13891 doi: 10.1038/ncomms13891 (2016).

This python code is responsible for randomly placing AQP expressing cells inside 
of an extended fcc lattice and then simulating diffusion using a Monte Carlo
random walk. Water molecules are randomly initialized in the lattice at the first time step. 
Water square displacements from original position are saved as vectors to a text file. 

Usage instructions:

Use the matlab script available on the Shapiro Lab website to calculate ADC for the saved diffusion path. 

"""
import numpy
from multiprocessing import Pool
"""Multiprocessing allows for parallel excecution of a specific function in python. 
#The degree of parallelism is limited to the nubmer of processors on your machine, 
#but as the diffusion problem is trivially parallelized, Excecuting this program 
#on mutliple machines simultaneously on an HPC cluster allows for significant 
#speed-up over serial computing."""
lattice=numpy.array([[1,0,0], [0,1,0],[0,0,1],[-1,0,0],[0,-1,0],[0,0,-1],[1,1,1],[1,1,-1],[1,-1,1],[1,-1,-1],[-1,1,1],[-1,1,-1],\
[-1,-1,-1],\
[-1,-1,1],\
[-2,-2,-1],\
[-2,-2,1],\
[-2,-1,-2],\
[-2,-1,0],\
[-2,-1,2],\
[-2,0,-1],\
[-2,0,1],\
[-2,1,-2],\
[-2,1,0],\
[-2,1,2],\
[-2,2,-1],\
[-2,2,1],\
[-1,-2,-2],\
[-1,-2,0],\
[-1,-2,2],\
[-1,0,-2],\
[-1,0,2],\
[-1,2,-2],\
[-1,2,0],\
[-1,2,2],\
[0,-2,-1],\
[0,-2,1],\
[0,-1,-2],\
[0,-1,2],\
[0,1,-2],\
[0,1,2],\
[0,2,-1],\
[0,2,1],\
[1,-2,-2],\
[1,-2,0],\
[1,-2,2],\
[1,0,-2],\
[1,0,2],\
[1,2,-2],\
[1,2,0],\
[1,2,2],\
[2,-2,-1],\
[2,-2,1],\
[2,-1,-2],\
[2,-1,0],\
[2,-1,2],\
[2,0,-1],\
[2,0,1],\
[2,1,-2],\
[2,1,0],\
[2,1,2],\
[2,2,-1],\
[2,2,1],\
[-2,-2,3],\
[-2,-2,-3],\
[-2,0,3],\
[-2,0,-3],\
[-2,2,3],\
[-2,2,-3],\
[-2,3,-2],\
[-2,-3,-2],\
[-2,3,0],\
[-2,-3,0],\
[-2,3,2],\
[-2,-3,2],\
[-1,-1,3],\
[-1,-1,-3],\
[-1,1,3],\
[-1,1,-3],\
[-1,3,-1],\
[-1,-3,-1],\
[-1,3,1],\
[-1,-3,1],\
[0,-2,3],\
[0,-2,-3],\
[0,0,3],\
[0,0,-3],\
[0,2,3],\
[0,2,-3],\
[0,3,-2],\
[0,-3,-2],\
[0,3,0],\
[0,-3,0],\
[0,3,2],\
[0,-3,2],\
[1,-1,3],\
[1,-1,-3],\
[1,1,3],\
[1,1,-3],\
[1,3,-1],\
[1,-3,-1],\
[1,3,1],\
[1,-3,1],\
[2,-2,3],\
[2,-2,-3],\
[2,0,3],\
[2,0,-3],\
[2,2,3],\
[2,2,-3],\
[2,3,-2],\
[2,-3,-2],\
[2,3,0],\
[2,-3,0],\
[2,3,2],\
[2,-3,2],\
[3,-2,-2],\
[-3,-2,-2],\
[3,-2,0],\
[-3,-2,0],\
[3,-2,2],\
[-3,-2,2],\
[3,-1,-1],\
[-3,-1,-1],\
[3,-1,1],\
[-3,-1,1],\
[3,0,-2],\
[-3,0,-2],\
[3,0,0],\
[-3,0,0],\
[3,0,2],\
[-3,0,2],\
[3,1,-1],\
[-3,1,-1],\
[3,1,1],\
[-3,1,1],\
[3,2,-2],\
[-3,2,-2],\
[3,2,0],\
[-3,2,0],\
[3,2,2],\
[-3,2,2],\
[1,-3,-3],\
[1,-3,3],\
[1,3,-3],\
[1,3,3],\
[-1,-3,-3],\
[-1,-3,3],\
[-1,3,-3],\
[-1,3,3],\
[3,-1,-3],\
[3,-1,3],\
[-3,-1,-3],\
[-3,-1,3],\
[3,1,-3],\
[3,1,3],\
[-3,1,-3],\
[-3,1,3],\
[3,-3,-1],\
[3,3,-1],\
[-3,-3,-1],\
[-3,3,-1],\
[3,-3,1],\
[3,3,1],\
[-3,-3,1],\
[-3,3,1],\
[-3,-3,-3],\
[-3,-3,3],\
[-3,3,-3],\
[-3,3,3],\
[3,-3,-3],\
[3,-3,3],\
[3,3,-3],\
[3,3,3]])

particlexyz=numpy.zeros((160,18000,3))
x=list([0])
for dd in x:
    particlenumber=4; "Number of water molecules in simulation"

    tau = .05;

    
    def diffusionsim(xx):
        numpy.random.seed(xx)
        radius=6.8+numpy.random.randn(1)*1.25; "Adjust Cell Radius and population variability as needed"
        box_size=2*numpy.sqrt(2)*radius; "Size of FCC box. Do not adjust."
        t =18000; "Number of Time steps."
        particle_xyz=numpy.zeros([t,3]); "Initialize water molecule."
        particle_xyz[0,:]=box_size*(numpy.random.rand(1,3)-.5);
        xcount=numpy.zeros([t,1]);
        P_Expressing=.25; "Permeability in micron per ms of  expressing cells"
        P_Control=.036; "Permeability in micron per ms of non-expressing cells"
        tau = .05;
        D_Cellular=.65; "D in micron^2 per ms inside the cell"
        D_ExtraCellular=2.1; "D in micron^2 per ms outside the cell"
        centers=numpy.sqrt(2)*radius*lattice; "Center position of each cell in the lattice. Do not change."
        aqp_Reflec=1-(tau/(6*D_Cellular))**(.5)*4*P_Expressing; "Reflection Probabilities calculated from membrane permeability"
        Basal_Reflec=1-(tau/(6*D_Cellular))**(.5)*4*P_Control; "Reflection probability for non-expressing cells in lattice."
        expr=numpy.concatenate((numpy.ones((dd,1)),numpy.zeros((108-(dd),1))),axis=0); "Populate a vector with 1's for expressing and 0's for non-expressing cells."
        ind=numpy.random.permutation(108); "Randomly permute expr to ensure random spatial arrangement of expressing cells."
        rand_expr1=expr[ind]; "The following lines ensure rational distribution of expressing cells at the periodic boundary. (That is, they ensure traveling through the periodic boundary does not change the expr status of the resident cell."
        wholeexpression=rand_expr1[0:62]
        halfexpression=rand_expr1[62:101]
        quarterexpression=rand_expr1[101:107]
        eighthexpression=rand_expr1[107]
        
        randexpr=numpy.concatenate((wholeexpression[:,0],halfexpression[:,0],halfexpression[:,0],\
        quarterexpression[:,0],quarterexpression[:,0],quarterexpression[:,0],quarterexpression[:,0],\
        eighthexpression,eighthexpression,eighthexpression,eighthexpression,eighthexpression,eighthexpression,eighthexpression,eighthexpression),axis=0)
        rand_expr=randexpr[:,None]
        R_io=aqp_Reflec*rand_expr+Basal_Reflec*(numpy.ones((172,1))-rand_expr);
        R_oi=1-((1-R_io)*numpy.sqrt(D_Cellular/D_ExtraCellular));
        "Starts the for loop for diffusion below."
        for j in list(numpy.array(range(t-1))+1):
            Xr=numpy.random.rand(1);
            q=numpy.argmin(numpy.sum((particle_xyz[j-1,:]-centers)**2,axis=1))
            bk=centers[q,:];       
            if numpy.sum(numpy.square(particle_xyz[j-1,:]-bk))<radius**2:
                particle_xyz[j,:]=particle_xyz[j-1,:]+numpy.random.randn(1,3)*numpy.sqrt(numpy.pi/2)*numpy.sqrt(2*D_Cellular*tau);
            else:
                particle_xyz[j,:]=numpy.squeeze(particle_xyz[j-1,:])+numpy.random.randn(1,3)*numpy.sqrt(numpy.pi/2)*numpy.sqrt(2*D_ExtraCellular*tau);            
            
            if (numpy.sum(numpy.square(particle_xyz[j-1,:]-bk))<radius**2 and numpy.sum(numpy.square(particle_xyz[j,:]-bk))>radius**2 and Xr<R_io[q]) or (numpy.sum(numpy.square(particle_xyz[j-1,:]-bk))>radius**2 and numpy.sum(numpy.square(particle_xyz[j,:]-bk))<(radius)**2 and Xr<R_oi[q]):
                particle_xyz[j,:]=particle_xyz[j-1,:];        
            xcount[j]=xcount[j-1];
            if particle_xyz[j,0]>3*box_size/2: 
               particle_xyz[j,0]=particle_xyz[j,0]-3*box_size;
               xcount[j]=xcount[j]+1;
            
            if particle_xyz[j,0]<-3*box_size/2:
               particle_xyz[j,0]=particle_xyz[j,0]+3*box_size;
               xcount[j]=xcount[j]-1;
            
            if particle_xyz[j,2]>3*box_size/2 or particle_xyz[j,2]<-3*box_size/2:
               particle_xyz[j,2]=particle_xyz[j,2]-numpy.sign(particle_xyz[j,2])*3*box_size;              
            
            if particle_xyz[j,1]>3*box_size/2 or particle_xyz[j,1]<-3*box_size/2:
               particle_xyz[j,1]=particle_xyz[j,1]-numpy.sign(particle_xyz[j,1])*3*box_size;
      
        particle_xyz[:,0]=particle_xyz[:,0]+numpy.transpose(numpy.array((xcount)))*3*box_size;
        
        return particle_xyz
       
    
    def ADCcalc():
        Delta_Index=numpy.array(range(801))
        Delta_Steps=numpy.shape(Delta_Index)[0]
        Delta=numpy.array(Delta_Index)/tau
        
        for p in range(Delta_Steps):
            Delta[p]=numpy.floor(Delta[p])
        sqdist=numpy.array(numpy.zeros([Delta_Steps,particlenumber]))
        for p in list(numpy.array(range(Delta_Steps-1))+1):
            sqdist[p,:]=numpy.square(particlexyz[:,int(Delta[p]),0]-particlexyz[:,0,0])
        return sqdist
    
    pool = Pool(4)
    particlexyz=pool.map(diffusionsim,numpy.random.randint(1,high=10000000,size=particlenumber)) "Random seed for parallel for loops"
    particlexyz=numpy.array(particlexyz)
    results=ADCcalc()
    print(numpy.sqrt(results))
    numpy.savetxt(str(dd)+'unclat'+str(numpy.random.randint(10000))+'.out', results, fmt='%.18e', delimiter=' ', newline='\n', header='', footer='', comments='# ') "Save output array with random number flag to distinguish from parallel machine runs."
